/* 
 * ASCD : copyright (c) Aley Keprt 1998-2001
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File mem.c - where memory is actually held.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*
 * NOTES
 * -----
 * The SAM memory is implemented in NMAINPAGES 16K pages and pointed to 
 * by the array *realMem[]. (The actual space is allocated at run time)
 * The Z80 Address space is asigned using an array of 4 pointers to this
 * space. The page number asigned to Sections A..D of the address space
 * is stored in real_pagenos[].
 * All memory addressing in the Z80 emulation is done indirectly via these
 * pointers - see file mem.h
 */


#include <stdlib.h>
#include <stdio.h>

#include "mem.h"
#include "util.h"
#include "resource.h"
#include "audio.h"
#include "zlib.h"
#include "kbd.h"
#include "z80.h"


uns8 *theRdMemory[4];		/* 4 pointers to real memory */
uns8 *theWrMemory[4];		/* 4 pointers to real memory */
int real_pagenos[4];		/* index to real memory page */
uns8 *realMem[TOTAL_NO_PAGES];	/* real memory is kept here */
uns8 *scratchMem[4];            /* scratch memory for read only areas */


extern int toplines;
extern int totallines;


char exe_path[1000];


//---------------------------------------------------------------------------


void init_memory(void) {
//
//allocate NMAINPAGES chunks of main memory, and the required
//number of external memory pages, each of size SAM_PAGESIZE
//
	int i,iMeg;

  if(realMem[0]) return; //already allocated

	/* initialize the main SAM memory */
	for(i=0;i<NMAINPAGES;i++) {
		if((realMem[i] = (uns8 *)malloc(SAM_PAGESIZE)) == (uns8 *)0)
			simcoupe_mesg(SIMCOUPE_FATAL, "can't allocate 16K for page %d!", i);
	}

	/* initialize the memory for the 2 roms */
	for(i=0;i<2;i++) {
		if((realMem[ROM0+i] = (uns8 *)malloc(SAM_PAGESIZE)) == (uns8 *)0)
			simcoupe_mesg(SIMCOUPE_FATAL, "can't allocate 16K for ROM %d!", i);
	}

	/* initialize the memory for the 4 scratch pages */
	for(i=0;i<4;i++) {
		if((scratchMem[i] = (uns8 *)malloc(SAM_PAGESIZE)) == (uns8 *)0)
			simcoupe_mesg(SIMCOUPE_FATAL, "can't allocate 16K for scratch area %d!", i);
	}

	/* initialize the external memory if required */
	if(cfg.extmem>MAXEXPAGES) cfg.extmem=MAXEXPAGES;
	for(iMeg=0;iMeg<cfg.extmem;iMeg++) {
    for(i=0;i<NEXPAGES;i++) {
      if(!(realMem[NMAINPAGES+i+(3-iMeg)*NEXPAGES]=(uns8 *)malloc(SAM_PAGESIZE))) {
				simcoupe_mesg(SIMCOUPE_FATAL, "can't allocate 16K for external page %d!", i+iMeg*NEXPAGES);
			}
		}
	}
}


//---------------------------------------------------------------------------


void get_exe_path(const char *exe) {
//
//get exe's path from the argv[0]
//
 char *p1=strrchr(exe,'/');
 char *p2=strrchr(exe,'\\');
 char *p3=strrchr(exe,':');

 if(!p1) p1=p3;
 if(!p2) p2=p3;

 if(p1>p3) p3=p1;
 if(p2>p3) p3=p2;

 strcpy(exe_path,exe);
 exe_path[p3-exe]=0;
}


//---------------------------------------------------------------------------


static void hookrom(int page,int pos) {
  realMem[page][pos+0]=0xed; //break (2 bytes)
  realMem[page][pos+1]=0xfe;
  realMem[page][pos+2]=201;  //ret
}


//---------------------------------------------------------------------------


int load_rom_image(char *image, int page) {
//
//read ROM image "image" into real memory page <page>
//
//tries to open the file in the current directory, then in the exe's directory
//all files are readed via zlib
//
//supports .gz files
//
  gzFile *f;
  int nread;

  //is where to load?
  if(!realMem[page]) {
    simcoupe_mesg(SIMCOUPE_ERR,"Can't load \"%s\" - memory isn't allocated",image);
    return;
  }

  //try to load it directly
  f=gzopen(image,"rb");
  if(!f) {
    //try to load it from exe location
    char fname[1000];
    sprintf(fname,"%s\\%s",exe_path,image);
    f=gzopen(fname,"rb");
    if(!f) {
      simcoupe_mesg(SIMCOUPE_ERR,"Can't open ROM image \"%s\"",image);
      return -1;
    }
  }

  nread=gzread(f,realMem[page],SAM_PAGESIZE);

  //give a warning if image is not 16k
  if(nread!=SAM_PAGESIZE) {
    simcoupe_mesg(SIMCOUPE_WARN,"ROM image \"%s\" is only %d bytes",image,nread);
  }

  gzclose(f);

  //tape-loading trap
  switch(cfg.zxmode) {
  case EM_128k:
    if(page!=ROM1) break;
  case EM_48k:
    hookrom(page,0x56c); //load/verify
    hookrom(page,0x5ba); //mr.copy+ re-entry
    hookrom(page,0x4d8); //save
    break;

  case EM_Sam:
    if(page!=ROM1) break;
    hookrom(page,0xe679&0x3fff); //load/verify
    hookrom(page,0xe56a&0x3fff); //save
    break;
  }

  return 0;
}


//---------------------------------------------------------------------------


inline void page_in(int bank, int page, int readonly) {
//
//page in real memory page at bank 0-3
//
  theRdMemory[bank]=realMem[page];
  if(!readonly) {
    theWrMemory[bank]=realMem[page];
    real_pagenos[bank]=page;
  } else {
    theWrMemory[bank]=scratchMem[bank];
    real_pagenos[bank]=SCRATCH_PAGE;
  }
}


//---------------------------------------------------------------------------


void set_emulation_mode(EM mode) {
//
//exit zx mode, i.e. load ZXS ROM, set palette, memory pages, etc.
//
//mode: 0=sam, 1=48k, 2=128k
//
  int i;
  static Z80_Regs regs;
  static int curmode=-1;

  if(!realMem[0]) return; //memory isn't allocated yet

  //never set the same mode again - this would cause reset!
  if(mode==curmode) return;
  curmode=mode;
/*
  simcoupe_mesg(SIMCOUPE_INFO,"Switching to mode %i",mode);
*/
  init_ports();

  page_in(0,ROM0,True);
  page_in(1,5,False);
  page_in(2,2,False);
  page_in(3,0,False);
  out_zxpage(0);

  set_zx_palette();

  //reset SAA1099
  for(i=31;i>=0;i--) {
    saa1099_control_port_w(0,0,i);
    saa1099_write_port_w(0,0,0);
  }

  //reset AY8910
  for(i=0;i<16;i++) {
    AY8910Write(0,i,0);
  }

  cfg.zxmode=mode;

  switch(mode) {

  case EM_48k:
    SpectrumKeyboard=1;

    load_rom_image("spectrum.rom",ROM0);

    toplines=64;
    totallines=312;
    break;

  case EM_128k:
    SpectrumKeyboard=1;

    load_rom_image("zx128_0.rom",ROM0);
    load_rom_image("zx128_1.rom",ROM1);

    toplines=63;
    totallines=311;
    break;

  case EM_Sam:
    page_in(0,ROM0,True);
    page_in(1,1,False);
    page_in(2,2,False);
    page_in(3,3,False);

    load_rom_image("sam_rom0.rom",ROM0);
    load_rom_image("sam_rom1.rom",ROM1);
    SpectrumKeyboard=0;

    toplines=68; //??? maybe 69
    totallines=312;
    break;
  }

  if(cfg.pallines) totallines=cfg.pallines;

  AudioReinit();

  set_registers(&regs);
}

