/*
 * ASCD : copyright (c) Aley Keprt 1998-2001
 *
 * File z80_snap.c - 48/128k .z80 file handling
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include "zlib.h"

#include "z80.h"
#include "z80_snap.h"
#include "mem.h"


typedef unsigned char byte;


extern byte *realMem[];


typedef struct {
  byte a,f;
  byte bc[2],hl[2],pc[2],sp[2];
  byte i,r;  //r bit 7 not signiicant!
  byte flags;
/*
    Bit 0  : Bit 7 of the R-register
    Bit 1-3: Border colour
    Bit 4  : 1=Basic SamRom switched in
    Bit 5  : 1=Block of data is compressed
*/
  byte de[2],bc1[2],de1[2],hl1[2];
  byte a1,f1;
  byte iy[2],ix[2];
  byte iff1,iff2;
  byte flags2;
/*
    Bit 0-1: Interrupt mode (0, 1 or 2)
    Bit 2  : 1=Issue 2 emulation
    Bit 3  : 1=Double interrupt frequency
    Bit 4-5: 1=High video synchronisation
             3=Low video synchronisation
             0,2=Normal
    Bit 6-7: 0=Cursor/Protek/AGF joystick
             1=Kempston joystick
             2=Sinclair 2 Left joystick (or user
               defined, for version 3 .Z80 files)
             3=Sinclair 2 Right joystick
*/
} Z80_header;


//-----------------------------------------------------------------------------------


static void unpack(byte *w,byte *r,int left) {
//
//.z80 unpack
//
// w    = write pointer
// r    = read pointer
// left = unpacked size
//
  while(left>0) {
    if(r[0]!=0xed || r[1]!=0xed) {
      *w++=*r++;
      left--;
    } else {
      int num=r[2];
      int v=r[3];
      r+=4;
      if(num>left) num=left; //protect against memory overflow
      left-=num;
      do *w++=v; while(--num);
    }
  }
}


//-----------------------------------------------------------------------------------


#define mw(reg) (r.reg=hdr.reg[0]+hdr.reg[1]*256)
#define mb(h,l,hl) (r.h=hdr.hl[1],r.l=hdr.hl[0])


int _load_z80(gzFile *f) {
//
//internal function
//load .z80 snapshot
//
//return: emulation mode (0=error, 1=48k, 2=128k)
//
  Z80_header hdr;
  Z80_Regs r;
  int mode128=0;

  struct Z80_Header_2 {
    byte size[2],pc[2];
    byte hw;
    byte page;
  } hdr2;

  if(gzread(f,&hdr,30)!=30) return 0; //error
  if(hdr.flags==255) hdr.flags=1;

  //initialize registers
  r.a=hdr.a;
  r.f=hdr.f;

  mb(b,c,bc);
  mb(d,e,de);
  mb(h,l,hl);

  mw(ix);
  mw(iy);
  mw(sp);
  mw(pc);

  r.a1=hdr.a1;
  r.f1=hdr.f1;

  mb(b1,c1,bc1);
  mb(d1,e1,de1);
  mb(h1,l1,hl1);

  r.iff1=(hdr.iff1!=0);
  r.iff2=(hdr.iff2!=0);

  r.i=hdr.i;
  r.r=(hdr.r&0x7f)+(hdr.flags&1)*128;

  r.im=hdr.flags2&3;
  if(r.im>=1) r.im++; //we have im0123, .z80 has im0x12!!!

  //read extended header, if present
  if(r.pc==0) {

    //new .z80 format (128k)
    int size;

    if(gzread(f,&hdr2,sizeof(hdr2))!=sizeof(hdr2)) return 0; //error

    size=hdr2.size[0]+hdr2.size[1]*256;
    r.pc=hdr2.pc[0]+256*hdr2.pc[1];

    if(size==23) {
      if(hdr2.hw==3) mode128=1;
    } else {
      if(hdr2.hw==5 || hdr2.hw==6) mode128=1;
    }
    if(hdr2.hw==4) mode128=1;

    //skip the rest of the header
//    gzseek(f,size,SEEK_CUR);
    while(size>sizeof(hdr2)-2) gzgetc(f),size--;

    //read extended body
    while(1) {
      byte *buf;
      int page;

      size=gzgetc(f);
      size+=gzgetc(f)*256;
      page=gzgetc(f);
      if(gzeof(f)) break;

      buf=(byte*)malloc(size);
      if(gzread(f,buf,size)!=size) return 0; //error

      if(mode128) {
        //128k verze - stranky jsou cislovany 3-10
        unpack(realMem[page-3],buf,16384);
      } else {
        //48k verze - stranky jsou cislovany 8,4,5
        switch(page) {
        case 8: unpack(realMem[5],buf,16384); break;
        case 4: unpack(realMem[2],buf,16384); break;
        case 5: unpack(realMem[0],buf,16384); break;
        }
      }

      free(buf);
    }

  } else {

    //old .z80 format (48k)

    if(hdr.flags&32) {
      //compressed body
      byte *buf=(byte*)malloc(49152);
      byte *body=(byte*)malloc(49152);

      if(gzread(f,buf,49152)) {
        unpack(body,buf,49152);

        memcpy(realMem[5],&body[0],16384);
        memcpy(realMem[2],&body[16384],16384);
        memcpy(realMem[0],&body[32768],16384);
      } //else disk read error
      free(buf);
      free(body);
    } else {
      //plain body
      if(!gzread(f,realMem[5],16384)) return 0; //error
      if(!gzread(f,realMem[2],16384)) return 0; //error
      if(!gzread(f,realMem[0],16384)) return 0; //error
    }
  }

  //initialize ports
  out_byte(BORDER_PORT,hdr.flags>>1 & 7);

  set_emulation_mode(mode128 ? EM_128k : EM_48k);
  set_registers(&r);
  if(mode128) out_zxpage(hdr2.page);

  return 1+mode128;
}
#undef mw
#undef mb


//-----------------------------------------------------------------------


#define mw(reg) (hdr.reg[0]=r.reg%256, hdr.reg[1]=r.reg/256)
#define mb(h,l,hl) (hdr.hl[1]=r.h, hdr.hl[0]=r.l)


int save_z80(const char *fname) {

 Z80_header hdr;
 Z80_Regs r;
/*
 byte dump[49152];

 int dia=dialog_running;
 HaltDialog();

 get_registers(&r);

 //store registers
 mw(ix);
 mw(iy);
 r.sp-=2;
 mw(sp);

 mb(a,f,af);
 mb(b,c,bc);
 mb(d,e,de);
 mb(h,l,hl);

 mb(a1,f1,af1);
 mb(b1,c1,bc1);
 mb(d1,e1,de1);
 mb(h1,l1,hl1);

 sna.im=r.im;
 if(sna.im>=2) sna.im--; //my mame im0123, sna ma im012!!!
 sna.r=r.r;
 sna.i=r.i;

 sna.iff=r.iff1+r.iff2*2;

 sna.border=border&7;

 write_word(r.sp,r.pc);

 memcpy(&dump[0],realMem[1],16384);
 memcpy(&dump[16384],realMem[2],16384);
 memcpy(&dump[32768],realMem[3],16384);

 {
/*
   gzFile *f=gzopen(fname,"wb");
   if(gzwrite(f,&sna,27)!=27) return -1; //error
   if(gzwrite(f,dump,49152)!=49152) return -1; //error
   gzclose(f);

   FILE *f=fopen(fname,"wb");
   fwrite(&sna,27,1,f);
   fwrite(dump,49152,1,f);
   fclose(f);
 }

 if(dia) ResumeDialog();
*/
}
#undef mw
#undef mb


//-----------------------------------------------------------------------


int load_z80(const char *fname) {
//
//load .z80 snapshot
//
//return: error flag
//
  gzFile *gzf;
  int r;

  gzf=gzopen(fname,"rb");
  if(!gzf) return -1;

  r=_load_z80(gzf); //returns emulation mode
  gzclose(gzf);
  return !r; //returns error flag
}

